## **Browsing the DA 2.0 address space**

For **Data Access 2.0** OPC Servers the browsing mechanism is implemented on **DaAddressSpaceElement** object. When browsing, a DaAddressSpaceElement array is returned and it contains all "children" elements of the browsed DaAddressSpaceElement. Each returned DaAddressSpaceElement object is uniquely identified in the server's address space by its **ItemId**.

The result of browsing is hierarchical. Starting from the address space element specified by its identifier, the client will get information on the nodes and leaves below. With **DaAddressSpaceElementBrowseOptions** parameter of the Browse method, the client can specify what he wants to get:

-   only nodes below the current node
    
-   only leaves below the current node
    
-   everything below the current node (all the nodes and leaves below the current node)
    

It is also possible to browse an OPC DA 2.0 server`s namespace using DaSession.Browse API method, although not recommended. If the DA2.0 server does not fully implement the browse functionality, that is if the OPC DA 2.0 server does not implement the OPC_BROWSE_TO browse direction, the DaAddressSpaceElement.Browse API method must be used.

### Note
```
It is recomended to always use the DaAddressSpaceElement.Browse API method for browsing OPC DA servers which 
may implement various OPC DA specifications. This is a single-method approach, no matter what OPC DA specification 
is implemented by the OPC DA server (and whether the respective OPC DA specification is fully implemented or not).
```
When browsing OPC DA servers through the DaAddressSpaceElement.Browse API method, it is mandatory that only the root node is manually constructed (with id "" and handle 0) and no other node element (with id different than "" and/or handle different than 0) because it would raise the case when more DaAddressSpaceElement API objects point to the same Toolkit CORE object.
### Note
```
The destruction of two API objects with the same handle would cause the Toolkit CORE to try and dispose the same 
object twice (which is not allowed inside the Toolkit CORE). Once the first API object was destroyed (thus having the 
corresponding Toolkit CORE object being disposed) the second API object will no longer point to a valid Toolkit CORE 
object, hence there should never be two API objects with the same handle. The Toolkit CORE may later create 
another object with the same handle as the one inside the "clone" API object thus causing unpredicted behavior.
```
### Example

The following example shows how to **synchronously** browse all the node and leaves of a virtual root of the _Softing OPC Toolkit OPC DA Demo Server_. If there is necessary to browse other levels, extend the following sample to call the **browse** method for each node obtained from the previous call of the browse method. For more detailed example on this topic have a look on the **DABrowse_Da20** sample.
### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...

// gets the application instance
Application application = Application.Instance;

// initializes the application
application.Initialize();

// creates a new DaSession object and adds it to the application
DaSession session = new DaSession("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}");

// sets the execution options
ExecutionOptions executionOptions = new ExecutionOptions();
executionOptions.ExecutionType = EnumExecutionType.SYNCHRONOUS;
executionOptions.ExecutionContext = 0;

// connects object to the server
session.Connect(true, false, executionOptions);

// obtains the root address space element
DaAddressSpaceElement rootDaAddressSpaceElement = new DaAddressSpaceElement(EnumAddressSpaceElementType.BRANCH, String.Empty, string.Empty, string.Empty, 0, null);
rootDaAddressSpaceElement.Session = session;        

DaAddressSpaceElement[] addressSpaceElements = null;
DaAddressSpaceElementBrowseOptions browseOptions = new DaAddressSpaceElementBrowseOptions();

// browses all the branches of the current node
browseOptions.ElementTypeFilter = EnumAddressSpaceElementType.BRANCH;

if (ResultCode.SUCCEEDED(rootDaAddressSpaceElement.Browse(
        browseOptions,
        out addressSpaceElements,
        executionOptions)))
{
    // all branches of current node are in the addressSpaceElements array
    ...
}

addressSpaceElements = null;

// browses all the leaves of the current node
browseOptions.ElementTypeFilter = EnumAddressSpaceElementType.LEAF;

if (ResultCode.SUCCEEDED(rootDaAddressSpaceElement.Browse(
        browseOptions,
        out addressSpaceElements,
        executionOptions)))
{
    // all leaves of current node are in the addressSpaceElements array
    ...
}

addressSpaceElements = null;

// gets all the properties of the current element
DaProperty[] daProperties = null;
DaGetPropertiesOptions propertyGetOptions = new DaGetPropertiesOptions();
propertyGetOptions.WhatPropertyData = EnumPropertyData.NONE;

if (ResultCode.SUCCEEDED(rootDaAddressSpaceElement.GetDaProperties(                        
        propertyGetOptions,
        out daProperties, 
        executionOptions)))
{
    // all properties of current node are in the allProperties array
    ...
}  

daProperties = null;

// terminates the application
application.Terminate();

session = null;
```
### VB .NET
```
Imports Softing.OPCToolbox.Client
Imports Softing.OPCToolbox
...

' gets the application instance
Dim application As Application = application.Instance

' initializes the application
application.Initialize()

' creates a new DaSession object and adds it to the application
Dim session As New DaSession("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}")

' sets the execution options
Dim executionOptions As ExecutionOptions = New ExecutionOptions()
executionOptions.ExecutionType = EnumExecutionType.SYNCHRONOUS
executionOptions.ExecutionContext = Convert.ToUInt32(0)

' connects object to the server
session.Connect(True, False, executionOptions)

' obtains the root address space element
Dim rootDaAddressSpaceElement As New DaAddressSpaceElement(EnumAddressSpaceElementType.BRANCH, String.Empty, String.Empty, String.Empty, 0, Nothing)
rootDaAddressSpaceElement.Session = session

Dim addressSpaceElements As DaAddressSpaceElement() = Nothing
Dim browseOptions As New DaAddressSpaceElementBrowseOptions

' browses all the branches of the current node
browseOptions.ElementTypeFilter = EnumAddressSpaceElementType.BRANCH

If ResultCode.SUCCEEDED(rootDaAddressSpaceElement.Browse(browseOptions, addressSpaceElements, executionOptions)) Then
    ' all branches of current node are in the addressSpaceElements array
    ...
End If

addressSpaceElements = Nothing

' browses all the leaves of the current node
browseOptions.ElementTypeFilter = EnumAddressSpaceElementType.LEAF

If ResultCode.SUCCEEDED(rootDaAddressSpaceElement.Browse(browseOptions, addressSpaceElements, executionOptions)) Then
    ' all leaves of current node are in the addressSpaceElements array
    ...
End If        

addressSpaceElements = Nothing

' gets all the properties of the current element
Dim daProperties As DaProperty() = Nothing
Dim propertyGetOptions As New DaGetPropertiesOptions
propertyGetOptions.WhatPropertyData = EnumPropertyData.NONE

If ResultCode.SUCCEEDED(rootDaAddressSpaceElement.GetDaProperties(propertyGetOptions, daProperties, executionOptions)) Then
    ' all properties of current node are in the allProperties array
    ...
End If

daProperties = Nothing

' terminates the application
application.Terminate()

session = Nothing
```
### C++
```
  using SoftingOPCToolboxClient;
  ...

  int _tmain(int argc, _TCHAR* argv[]) {

      // gets the application instance
      Application* application = getApplication();

      // initializes the application
      application->initialize();

      // creates a new DaSession object and adds it to the application
      DaSession* session = new DaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}"));

      // sets the execution options
      ExecutionOptions executionOptions;
      executionOptions.setExecutionType(EnumExecutionType_SYNCHRONOUS);
      executionOptions.setExecutionContext(0);

      // connects object to the server
      session->connect(TRUE, FALSE, &executionOptions);

      // obtains the root address space element
      tstring rootNodeText = session->getUrl();
      DaAddressSpaceElement* rootNode = new DaAddressSpaceElement(
          EnumAddressSpaceElementType_BRANCH, rootNodeText, _T(""), 0, _T(""), session);

      DaAddressSpaceElementBrowseOptions someBrowseOptions;
      std::vector<DaAddressSpaceElement*> addressSpaceElements;
      std::vector<DaAddressSpaceElement*>::iterator addressSpaceIterator;

      // browses all the branches of the current node
      someBrowseOptions.setElementTypeFilter(EnumAddressSpaceElementType_BRANCH);
      if (SUCCEEDED(rootNode->browse(&someBrowseOptions, addressSpaceElements, NULL)))
      {
          // all branches of current node are in the addressSpaceElements array
          ...
      }

      // deletes all address space elements
      for (addressSpaceIterator = addressSpaceElements.begin(); addressSpaceIterator != addressSpaceElements.end(); addressSpaceIterator++)
      {
          delete(*addressSpaceIterator);    
      }
      addressSpaceElements.clear();

      // browses all the leaves of the current node
      someBrowseOptions.setElementTypeFilter(EnumAddressSpaceElementType_LEAF);
      if (SUCCEEDED(rootNode->browse(&someBrowseOptions, addressSpaceElements, NULL)))
      {
          // all leaves of current node are in the addressSpaceElements array
          ...
      }

      // deletes all address space elements
      for (addressSpaceIterator = addressSpaceElements.begin(); addressSpaceIterator != addressSpaceElements.end(); addressSpaceIterator++)
      {
          delete(*addressSpaceIterator);
      }
      addressSpaceElements.clear();

      // gets all the properties of the current element
      std::vector<DaProperty*> allProperties;
      std::vector<DaProperty*>::iterator propertiesIterator;
      DaGetPropertiesOptions propertiesOptions;

      propertiesOptions.setWhatPropertyData(EnumPropertyData_ALL);
      if (SUCCEEDED(rootNode->getDaProperties(&propertiesOptions, allProperties, NULL)))
      {
          // all properties of current node are in the allProperties array
          ...
      }    

      // deletes all properties
      for (propertiesIterator = allProperties.begin(); propertiesIterator != allProperties.end(); propertiesIterator++)
      {
          delete(*propertiesIterator);
      }

      // terminates the application
      application->terminate();

      // release the resources used by application
      releaseApplication();

      delete session;
}
```
