## **Authenticated DA connection - reading and writing password protected items**

Reading and writing values is one of the most important task of a Data Access Client and also some servers provide the possibility to protect partially or totally the server address space with a specified password. In this case the client must be authenticated to be able to make operations on adress space items that are password protected.

The authentication is made for the current **DaSession** sample and the Data Access Client has the possibility to be authenticated or not. This can be done by using the following methods:

-   **DaSession::Logon Method** - Connects to a server using an user name and a password.
    
-   **DaSession::Logoff Method** - Ends the current logon session.
    

The authentication operation can be done either synchronous or asynchronous manner, based on how is most convenient.

With **synchronous** authentication, the Client calls the **Logon** or **Logoff** method and waits for the return values. For **asynchronous** manner the Client calls the authentication method but it does not wait until the authentication is actually done, being able to perform other interogations meanwhile.

###  Example 1

The following example shows how to **synchronously** authenticate a DA connection using DaSession. For a more detalied example on this topic check **DAConsole** example.

### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...

// gets the application instance
Application application = Application.Instance;

// initializes the application
application.Initialize();

// creates a new DaSession object and adds it to the application
DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

// sets the execution options
ExecutionOptions executionOptions = new ExecutionOptions();
executionOptions.ExecutionType = EnumExecutionType.SYNCHRONOUS;

// connects object to the server
session.Connect(true, false, executionOptions);

// login on secured DA connection
session.Logon("username", "password", null);

//  the secured connection is establish ... use it for read and write operations
...

// logout from secured DA connection
session.Logoff(null);

// terminates the application
application.Terminate();
```
### VB .NET
```
Imports Softing.OPCToolbox.Client
Imports Softing.OPCToolbox
...

' gets the application instance
Dim application As Application = Application.Instance

' initializes the application
application.Initialize()

' creates a new DaSession object and adds it to the application
Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

' sets the execution options
Dim executionOptions As New ExecutionOptions
executionOptions.ExecutionType = EnumExecutionType.SYNCHRONOUS

' connects object to the server
session.Connect(True, False, executionOptions)

' login on secured DA connection
session.logon("username", "password", Nothing)

' the secured connection is establish ... use it for read and write operations
...

' logout from secured DA connection
session.logoff(Nothing)

' terminates the application                
application.Terminate()
```
### C++
```
using namespace SoftingOPCToolboxClient;
...
int _tmain(int argc, _TCHAR* argv[]) 
{
    // gets the application instance
    Application* application = getApplication();

    // initializes the application
    application->initialize();

    // creates a new DaSession object and adds it to the application
    DaSession* session = new DaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}"));        

    // sets the execution options
    ExecutionOptions executionOptions;
    executionOptions.setExecutionType(EnumExecutionType_SYNCHRONOUS);

    // connects object to the server
    session->connect(TRUE, FALSE, &executionOptions);

    // login on secured DA connection
    session->logon(_T("username"), _T("password"), NULL);

    // the secured connection is establish ... use it for read and write operations
    ...

    // logout from secured DA connection
    session->logoff(NULL);

    // terminates the application
    application->terminate();
    // releases the resources used by application
    releaseApplication();

    delete session;
}
```

With **asynchronous** authentication, the client calls **Login** or **Logoff** method and then immediately gets feedback. After a certain interval of time, which depends on the kind of data access, the client gets the desired answer via a callback. When the authentication is complete in .NET the **LogonCompleted** event is raised and in C++ the **handleLogonCompleted** method is called. The same mechanism is implemented for logoff operation using **LogoffCompleted** event for .NET and **handleLogoffCompleted** method for C++ implementation.

###  Example 2

The following example shows how to **asynchronously** authenticate using DaSession object. For a more detailed example on this topic check the **DAConsole** example.

### C#
```
  class MainClass 
  {
    ...
    public static void HandleLogonCompleted(
        ObjectSpaceElement sender,
        uint executionContext,
        string userName,
        string password,
        int result)
    {
        if(ResultCode.SUCCEEDED(result))
        {
            // logon succeeded - do something
        }
        else
        {
             Console.WriteLine("Logon failed: " + result);
        } 
    }

    public static void HandleLogoffCompleted(
        ObjectSpaceElement sender,
        uint executionContext,
        int result)
    {
        if(ResultCode.SUCCEEDED(result))
        {
            // logoff succeeded - do something
        }
        else
        {
             Console.WriteLine("Logoff failed: " + result);
        } 
    }

    ...
    static void Main(string[] args) 
    {
        ...

        // gets the application instance
        Application application = Application.Instance;

        // initializes the application
        application.Initialize();

        // creates a new DaSession object and adds it to the application
        DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

        session.LogonCompleted += new LogonEventHandler(HandleLogonCompleted);
        session.LogoffCompleted += new LogoffEventHandler(HandleLogoffCompleted);

        // sets the execution options
        ExecutionOptions executionOptions = new ExecutionOptions();
        executionOptions.ExecutionType = EnumExecutionType.ASYNCHRONOUS;

        // connects object to the server
        session.Connect(true, false, executionOptions);

        // login on secured DA connection
        session.Logon("username", "password", executionOptions);

        executionOptions.ExecutionContext++;

        // the secured connection will be establish after the handleLogonCompleted callback
        // will be called ... use it for read and write operations
        ...

        // logout from secured DA connection
        session.Logoff(executionOptions);

        executionOptions.ExecutionContext++;
        ...

        application.Terminate();

    }
}
```
### VB .NET
```
Public Shared Sub handleLogonCompleted( _
    ByVal sender As ObjectSpaceElement, _
    ByVal executionContext As System.UInt32, _
    ByVal username As String, _
    ByVal password As String, _
    ByVal result As System.Int32)

    If ResultCode.SUCCEEDED(result) Then
         ' logon succeeded - do something
    Else
        System.Console.WriteLine("Logon failed: " + result.toString())
    End If
End Sub

Public Shared Sub handleLogoffCompleted( _
    ByVal sender As ObjectSpaceElement, _
    ByVal executionContext As System.UInt32, _
    ByVal result As System.Int32)

    If ResultCode.SUCCEEDED(result) Then
         ' logoff succeeded - do something
    Else
        System.Console.WriteLine("Logoff failed: " + result.toString())
    End If
End Sub
...
Shared Sub Main(ByVal args As String())
    ...

    ' gets the application instance
    Dim application As Application  = Application.Instance

    ' initializes the application
    application.Initialize()

    ' creates a new DaSession object and adds it to the application
    Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

    AddHandler session.LogonCompleted, AddressOf HandleLogonCompleted
    AddHandler session.LogoffCompleted, AddressOf HandleLogoffCompleted

    ' sets the execution options
    Dim executionOptions As New ExecutionOptions
    executionOptions.ExecutionType = EnumExecutionType.ASYNCHRONOUS

    ' connects object to the server
    session.Connect(True, False, executionOptions)

    ' login on secured DA connection
    session.logon("username", "password", executionOptions)

    ' increment execution context
    Decimal.op_Increment(Convert.ToDecimal(executionOptions.ExecutionContext))

    ' the secured connection will be establish after the handleLogonCompleted callback
    ' will be called ... use it for read and write operations
    ...

    ' logout from secured DA connection
    session.logoff(executionOptions)

    'increment execution context
    Decimal.op_Increment(Convert.ToDecimal(executionOptions.ExecutionContext))

    ...
    application.Terminate()

End Sub
```
### C++
```
using namespace SoftingOPCToolboxClient;
...

class MyDaSession : public DaSession{

public:

    MyDaSession (const tstring& url): DaSession(url) {}

    void handleLogonCompleted(
        unsigned long executionContext,
        tstring& userName,
        tstring& password,
        long result)
    {
        if (SUCCEEDED(result))
        {
            // logon succeeded - do something
        }
        else
        {
            _tprintf(_T("Logon failed [0x%8.8X]\n"), result);
        } 
    }

    void handleLogoffCompleted(
        unsigned long executionContext,
        long result)
    {
        if (SUCCEEDED(result))
        {
            // logoff succeeded - do something
        }
        else
        {
            _tprintf(_T("Logof failed [0x%8.8X]\n"), result);
        } 
    }
};

...
int _tmain(int argc, _TCHAR* argv[])
{
    // gets the application instance
    Application* application = getApplication();

    // initializes the application
    application->initialize();

    // creates a new DaSession object and adds it to the application
    MyDaSession* session = new MyDaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}"));    

    // sets the execution options
    ExecutionOptions executionOptions;
    executionOptions.setExecutionType(EnumExecutionType_ASYNCHRONOUS);
    executionOptions.setExecutionContext(0);

    // connects object to the server
    session->connect(TRUE, FALSE, &executionOptions);

    // login on secured DA connection
    session->logon(_T("username"), _T("password"), &executionOptions);

    unsigned long executionContext = executionOptions.getExecutionContext();
    executionOptions.setExecutionContext(++executionContext);

    // the secured connection will be establish after the handleLogonCompleted callback
    // will be called ... use it for read and write operations
    ...

    // logout from secured DA connection
    session->logoff(&executionOptions);

    executionOptions.setExecutionContext(++executionContext);

    application->terminate();    
    releaseApplication();

    delete session;
  }
```
