## **Value change notifications over datachanged event**



**Datachanged event** occurs when the client is notified with value changes within the items belonging to a subscription. This event is raised, at a time interval equal with the subscription's update rate, only if at least one of the items belonging to the subscription changes its value.

The server reads the values in cycles determined by the update rate and passes to the client via callback if a value change took place. When this callback is received from the server, the **Datachanged** event of the **DaSubscription** is raised (in .NET) and the **handleDataChange** method is called (in C++).

###  Example


The following example shows how to **asynchronously** read the values of four items using DaSubscription object and Datachanged events. For a more detalied example on this topic check **DAConsole** sample.
### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...

    public class MyDaSubscription : DaSubscription
{
    public MyDaSubscription(uint updateRate, DaSession parentSession) : base (updateRate, parentSession)
    {
        DataChanged += new DataChangedEventHandler(HandleDataChanged);
    }

    public static void HandleDataChanged(
        DaSubscription daSubscription, 
        DaItem[] items, 
        ValueQT[] values, 
        int[] results)
    {    

        for (int i = 0; i < results.Length; i++)
        {
            if(ResultCode.SUCCEEDED(results[i]))
            {    
                // data change received - use item and value
            }
            else
            {
                System.Console.WriteLine("Receiving item value failed " + results[i]);
            }
        }
    }
}

static void Main(string[] args) 
{
    // gets the application instance
    Application application = Application.Instance;

    // initializes the application
    application.Initialize();

    // creates a new DaSession object and adds it to the application
    DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

    // creates a new DaSubscription and add four items to the current subscription
    MyDaSubscription subscription = new MyDaSubscription(1000, session);

    DaItem daItem1 = new DaItem("maths.sin", subscription);
    DaItem daItem2 = new DaItem("time.local.second", subscription);
    DaItem daItem3 = new DaItem("increment.UI1", subscription);
    DaItem daItem4 = new DaItem("increment.BSTR", subscription);

      // sets the execution options
    ExecutionOptions executionOptions = new ExecutionOptions();
    executionOptions.ExecutionType = EnumExecutionType.SYNCHRONOUS;

    // connects object to the server
    session.Connect(true, false, executionOptions);

    // activate object to the server
    session.Connect(true, true, executionOptions);

    // session is activated use handleDataChanged to reads values
    ...

    // terminates the application    
    application.Terminate();

}
```
### VB .NET
```
Imports Softing.OPCToolbox.Client
Imports Softing.OPCToolbox
...

Public Class MyDaSubscription 
    Inherits DaSubscription

     Public Sub New(ByVal updateRate As UInt32, ByVal parentSession As DaSession)
        MyBase.New(updateRate, parentSession)
            AddHandler DataChanged, AddressOf HandleDataChanged

    End Sub

    Public Shared Sub HandleDataChanged( _
        ByVal daSubscription As DaSubscription, _
        ByVal items As DaItem(), _ 
        ByVal values As ValueQT(), _
        ByVal results As Integer())
        Dim i As Integer = 0

           While i < results.Length
            If ResultCode.SUCCEEDED(results(i)) Then
                ' data change received - use item and value
               Else
                System.Console.WriteLine("Datachange receive failed " + results(i).ToString())
            End If
            System.Math.Max(System.Threading.Interlocked.Increment(i), i - 1)
        End While
    End Sub
End Class

Shared Sub Main(ByVal args As String())
    ...

    ' gets the application instance
    Dim application As Application = application.Instance

    ' initializes the application
    application.Initialize()

    ' creates a new DaSession object and adds it to the application
    Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

    ' creates a new DaSubscription and add four items to the current subscription
    Dim subscription As New MyDaSubscription(Convert.ToUInt32(1000), session)

    Dim daItem1 As New DaItem("maths.sin", subscription)
    Dim daItem2 As New DaItem("time.local.second", subscription)
    Dim daItem3 As New DaItem("increment.UI1", subscription)
    Dim daItem4 As New DaItem("increment.BSTR", subscription)

    ' sets the execution options
    Dim executionOptions As New ExecutionOptions
    executionOptions.ExecutionType = EnumExecutionType.SYNCHRONOUS

    ' connects object to the server
    session.Connect(True, False, executionOptions)

    ' activate object to the server
    session.Connect(True, True, executionOptions)

    ' session is activated use handleDataChanged to reads values
    ...

    ' terminates the application
    application.Terminate()

End Sub
```
### C++
```
using namespace SoftingOPCToolboxClient;
...

class MyDaSubscription : public SoftingOPCToolboxClient::DaSubscription
{
    public:

    MyDaSubscription(unsigned long updateRate, DaSession* parentSession)
        : DaSubscription(updateRate, parentSession) {}

    void handleDataChanged(
            const std::vector<DaItem*>& items,
            const std::vector<ValueQT*>& values,
            const std::vector<long>& results)
    {
        for (unsigned long i = 0; i < results.size(); i++)
        {
            if (SUCCEEDED(results[i]))
            {
                // data change received - use item and value
            }
            else
            {
                _tprintf(_T("Datachange receive failed [0x%8.8X]\n"), results[i]);
            } 
        }    
    }

    virtual ~MyDaSubscription(){}
};

int _tmain(int argc, _TCHAR* argv[])
{
    // gets the application instance
    Application* application = getApplication();

    // initializes the application
    application->initialize();

    // creates a new DaSession object and adds it to the application
    DaSession* session = new DaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}"));

    // creates a new DaSubscription and add four items to the current subscription
    MyDaSubscription* subscription = new MyDaSubscription(1000, session);

    DaItem* item1 = new DaItem(_T("maths.sin"), subscription);
    DaItem* item2 = new DaItem(_T("time.local.second"), subscription);
    DaItem* item3 = new DaItem(_T("increment.UI1"), subscription);
    DaItem* item4 = new DaItem(_T("increment.BSTR"), subscription);

     // sets the execution options
    ExecutionOptions executionOptions;
    executionOptions.setExecutionType(EnumExecutionType_SYNCHRONOUS);
    executionOptions.setExecutionContext(0);

    // connect object to the server
    session->connect(TRUE, FALSE, &executionOptions);

    // activate object to the server
    session->connect(TRUE, TRUE, &executionOptions);

    // session is activated use handleDataChanged to reads values
    ...

    // terminates the application
    application->terminate();

    // releases the resources used by application
    releaseApplication();

    delete session;
}
```