## **Object states**

The **DaSession**, **AeSession**, **DaSubscription**, **AeSubscription** and **DaItem** objects are the equivalents in the Softing OPC Toolkit of the OPC objects. These objects are created by the client with the purpose of interacting with an OPC server.

The state of the connection of one of these objects with a server is called the **current state** of the object. In accordance with a client's needs, these objects can change their current state into another one, called **target state**.

Both current and target state can have a value from the ones listed below:

-   **Connected -** The object is connected to a server but no communication exists between these two.
    
-   **Activated -** The object is connected to a server and they communicate with each other via callbacks. When a DaSubscription object is in the activated state the client receives by means of callbacks the new values of the items that belong to the subscription. When an AeSubscription object is in the activated state the client receives by means of callbacks the events that occur in the server’s eventarea.
    
-   **Disconnected -** The object is not connected to a server.
    

The objects can move from one state to another (from current to target state), action named **state transition**. The state transition is illustrated in the figure below

![OPC-Classic-SDK](https://github.com/SoftingIndustrial/OPC-Classic-SDK/raw/main/documentation_pics/objectStates.PNG)

From a state an object can pass to any other state where changes from **activated** to **disconnected** and vice versa go through the **connected** state, what means each state change is reported by a connected event handler (.Net) or overridden method (C++).

Any state transition has two steps:

-   set the target state
    
-   perform the transition between current state and target state
    

Using the toolkit, the state transition can be realized using the **PerformStateTransition** method that is available for each object enumerated above.

The state transition can affect only the object itself or it can affect all the children of the object (in the hierarchical structure of the OPC object tree). If all the children of the object are affected we call this **deep** state transition. The first parameter of the method permits specifying if state transition is deep or not.

###  Example 1
The following example demonstrates how to perform a deep state transition for a DaSession object from disconnected (the initial state) to connected. The operation is done in a **synchronous** way.

For a more detailed example on this topic check the **DAConsole** example.
### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...
static void Main(string[] args) {
    ...
    // gets the application instance
    Application application = Application.Instance;
    application.Initialize();

    ExecutionOptions executionOptions = new ExecutionOptions();

    // creates a new DaSession object and adds it to the application
    DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

    // creates a new DaSubscription objects and adds it to the created session
    DaSubscription subscription = new DaSubscription(500, session);

    // creates a new DaItem object and adds it to the created subscription
    DaItem item1 = new DaItem("maths.sin", subscription);

    // sets the target state
    session.TargetState = EnumObjectState.CONNECTED;

    // performs deep state transition
    session.PerformStateTransition(true, executionOptions);

    ...

    // TODO disconnect session and remove items, subscription and session
    application.Terminate();
    ...
}
```
### VB .NET
```
Imports Softing.OPCToolbox.Client
Imports Softing.OPCToolbox
...
Shared Sub Main(ByVal args As String())
    ...
    'gets the application instance
    Dim application As Application = Application.Instance
    application.Initialize()

    Dim executionOptions As New ExecutionOptions

    'creates a new DaSession object and adds it to the application
    Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

    'creates a new DaSubscription objects and adds it to the created session
    Dim subscription As New DaSubscription ( Convert.ToUInt32(1000), session)

    'creates a new DaItem object and adds it to the created subscription
    Dim item1 As New DaItem("maths.sin" , subscription)

    'sets the target state
    session.TargetState = EnumObjectState.CONNECTED

    'performs deep state transition
    session.PerformStateTransition(True, executionOptions)

    'TODO disconnect session and remove items, subscription and session        
    application.Terminate()
    ...
End Sub
```
### C++
```
using namespace SoftingOPCToolboxClient;
...
int _tmain(int argc, _TCHAR* argv[])
{
    ExecutionOptions executionOptions;
    // gets the application instance
    Application* application = getApplication();

    application->initialize();

    // creates a new DaSession object and adds it to the application
    DaSession* session = new DaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}"));                        

    // creates a new DaSubscription objects and adds it to the created session            
    DaSubscription* subscription = new DaSubscription(1000, session);

    // creates a new DaItem object and adds it to the created subscription
    DaItem* item1 = new DaItem(_T("maths.sin"), subscription);

    // sets target state
    session->setTargetState(CONNECTED);

    // performs state transition; specifying NULL for execution options means the execution is synchronous
    session->performStateTransition(TRUE, &executionOptions);
    ...
    application->terminate();
    releaseApplication();

    delete session;
    delete subscription;
    delete item1;
}
```
### Example 2

The following example demonstrates how to perform a deep state transition for a DaSession object from disconnected(the initial state) to connected. The operation is done in an **asynchronous** way.

For a more detailed example on this topic check the **DAConsole** example.

### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...

public static void HandleStateChanged(ObjectSpaceElement sender, EnumObjectState state)
{
    if ((sender is DaSession) || (sender is DaSubscription))
    {                
        Console.WriteLine( sender + " State Changed - " + state);
    }
    if (sender is DaItem)
    {
        DaItem item = sender as DaItem;
        Console.WriteLine( item + " " + item.Id + " State Changed - " + state);
    }
}

public static void HandlePerformStateTransition(
    ObjectSpaceElement sender,
    uint executionContext,
    int result)
{            
    if(ResultCode.SUCCEEDED(result)){
        if ((sender is DaSession) || (sender is DaSubscription))
        {                
            Console.WriteLine( sender + " Performed state transition - "  + executionContext );
        }
        if (sender is DaItem)
        {
            DaItem item = sender as DaItem;
            Console.WriteLine( sender + " " + item.Id + " Performed state transition - "  + executionContext );
        }
    }
    else{
        Console.WriteLine("  Performed state transition failed");
    }
}
...
static void Main(string[] args) {

    // gets the application instance
    Application application = Application.Instance;

    application.Initialize();

    // set the execution options to asynchronous
    ExecutionOptions executionOptions = new ExecutionOptions();
    executionOptions.ExecutionType = EnumExecutionType.ASYNCHRONOUS;
    executionOptions.ExecutionContext = 0;

    // creates a new DaSession object and adds it to the application
    DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

    session.StateChangeCompleted += new StateChangeEventHandler(HandleStateChanged);
    session.PerformStateTransitionCompleted += new PerformStateTransitionEventHandler(HandlePerformStateTransition);

    // creates a new DaSubscription objects and adds it to the created session
    DaSubscription subscription = new DaSubscription(500, session);

    subscription.StateChangeCompleted += new StateChangeEventHandler(HandleStateChanged);
    subscription.PerformStateTransitionCompleted += new PerformStateTransitionEventHandler(HandlePerformStateTransition);

    // creates a new DaItem object and adds it to the created subscription
    DaItem item1 = new DaItem("maths.sin", subscription);

    item1.StateChangeCompleted += StateChangeEventHandler(HandleStateChanged);
    item1.PerformStateTransitionCompleted += new PerformStateTransitionEventHandler(HandlePerformStateTransition);

    // set the target state
    session.TargetState = EnumObjectState.CONNECTED;

    // perform deep state transition
    session.PerformStateTransition(true,executionOptions);
    executionOptions.ExecutionContext++;
    ...
} // end Main
```
### VB .NET
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...

Public Shared Sub HandleStateChanged(ByVal obj As ObjectSpaceElement, ByVal state As EnumObjectState)
    If Not Type.ReferenceEquals(obj.[GetType](), GetType(DaItem)) Then
        System.Console.WriteLine(Convert.ToString(obj) + " State Changed - " + Convert.ToString(state))
    Else
        Dim item As DaItem
        item = DirectCast(obj, DaItem)
        System.Console.WriteLine(Convert.ToString(obj) + " " + item.Id + " State Changed - " + Convert.ToString(state))
    End If
End Sub

Public Shared Sub HandlePerformStateTransition(ByVal obj As ObjectSpaceElement, ByVal executionContext As System.UInt32, ByVal result As System.Int32)
    If ResultCode.SUCCEEDED(result) Then
        If Not Type.ReferenceEquals(obj.[GetType](), GetType(DaItem)) Then
            System.Console.WriteLine(Convert.ToString(obj) + " Performed state transition - " + Convert.ToString(executionContext))
        Else
            Dim item As DaItem
            item = DirectCast(obj, DaItem)
            System.Console.WriteLine(Convert.ToString(obj) + " " + item.Id + " Performed state transition -" + Convert.ToString(executionContext))
        End If
    Else
        System.Console.WriteLine("  Performed state transition failed")
    End If
End Sub
...
Shared Sub Main(ByVal args As String())

    'gets the application instance
    Dim application As Application = Application.Instance

    application.Initialize()

    'set the execution options to asynchronous
    Dim executionOptions As New ExecutionOptions
    executionOptions.ExecutionType = EnumExecutionType.ASYNCHRONOUS
    executionOptions.ExecutionContext = Convert.ToUInt32(0)

    'creates a new DaSession object and adds it to the application
    Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

    AddHandler session.StateChangeCompleted, AddressOf HandleStateChanged
    AddHandler session.PerformStateTransitionCompleted, AddressOf HandlePerformStateTransition

    'creates a new DaSubscription objects and adds it to the created session
    Dim subscription As New DaSubscription ( Convert.ToUInt32(1000), session)

    AddHandler subscription.StateChangeCompleted, AddressOf HandleStateChanged
    AddHandler subscription.PerformStateTransitionCompleted, AddressOf HandlePerformStateTransition

    'creates a new DaItem object and adds it to the created subscription
    Dim item1 As New DaItem("maths.sin", subscription)

    AddHandler item1.StateChangeCompleted, AddressOf HandleStateChanged
    AddHandler item1.PerformStateTransitionCompleted, AddressOf HandlePerformStateTransition

    'set the target state
    session.TargetState = EnumObjectState.CONNECTED

    'perform deep state transition
    session.PerformStateTransition(True,executionOptions)
    'increment execution context
    Decimal.op_Increment(Convert.ToDecimal(executionOptions.ExecutionContext))
    ...
End Sub
```
### C++
```
using namespace SoftingOPCToolboxClient;
...
class MyDaSession : public DaSession{

    public:

        MyDaSession (const tstring& url): DaSession(url) {}
        void handleStateChangeCompleted(EnumObjectState state){

            _tprintf(_T("\n%s state changed - %s\n"), "Session",state);

        } // end handleStateChangeCompleted

        void handlePerformStateTransitionCompleted(
            unsigned long executionContext,
            long result)
        {
            _tprintf("\n");
            if (SUCCEEDED(result)){
                _tprintf(_T("%s performed state transition - context: +%lu\n"), "Session", executionContext);
            } // end if
            else{    
                _tprintf(" Performed state transition failed");
            }
        } // end handlePerformStateTransitionCompleted
} // end MyDaSession

int _tmain(int argc, _TCHAR* argv[])
{

    ExecutionOptions executionOptions;

    executionOptions.setExecutionType(ASYNCHRONOUS);
    executionOptions.setExecutionContext(0);
    executionContext = executionOptions.getExecutionContext();

    Application* application = getApplication();

    application->initialize();

    // creates a new DaSession object and adds it to the application
    MyDaSession* session = new MyDaSession("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}");    

    // creates a new DaSubscription object and adds it to the created session
    DaSubscription* subscription = new DaSubscription(1000, session);

    // creates a new DaItem object and adds it to the created subscription
    DaItem* item1 = new DaItem(_T("maths.sin"), subscription);

    // set target state
    session->setTargetState(CONNECTED);

    // perform state transition
    session->performStateTransition(TRUE, &executionOptions);}

    application->terminate();
    releaseApplication();

    delete session;
    delete subscription;
    delete item1;
}
```


All state transitions can be divided in three main categories:

-   The transition from any state to **connected**
    
-   The transition from any state to **activated**
    
-   The transition from any state to **disconnected**
    

The toolkit provides methods for all these state transitions. These methods are an alternative to setting the target state and then calling PerformStateTransition method. Instead of doing this, the **Connect** and **Disconnect** methods are used, with the parameters set according to the purpose of connection.

The Connect method permits specifying if the state transition is deep by setting the first parameter to true or false. Whether the object communicates or not with the server can be specified by setting the second parameter to true(the object is connected and communicates with the server ) or false( the object is connected but doesn’t communicate with the server).

The Disconnect method will disconnect the object and all his children from the server.

All the above enumerated methods can be executed synchronously or asynchronously. This is specified by the last parameter of these methods.

###  Example 3

The following example demonstrates how to deeply connect, activate and then disconnect a DaSession object. The operations are done in a **synchronous** way.

For a more detailed example on this topic check the **DAConsole** example.
### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;    
...
static void Main(string[] args)
{
    // gets the application instance
    Application application = Application.Instance;

    application.Initialize();

    ExecutionOptions executionOptions = new ExecutionOptions();

    // creates a new DaSession object and adds it to the application
    DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

    // creates a new DaSubscription object and adds it to the created session
    DaSubscription subscription = new DaSubscription(500, session);

    // creates a new DaItem object and adds it to the created subscription
    DaItem item1 = new DaItem("maths.sin", subscription);

    // connects session, subscription and item1 to the server
    session.Connect(true, false, executionOptions);

    // activates the session, subscription and item1      
    session.Connect(true, true, executionOptions);

    // disconnects session, subscription and item1 from the server
    session.Disconnect(executionOptions);

} // end Main
```
### VB .NET
```
Imports Softing.OPCToolbox.Client;
Imports Softing.OPCToolbox;
...
Shared Sub Main(ByVal args As String())
    ...
    'gets the application instance
    Dim application As Application = Application.Instance
    application.Initialize()

    Dim executionOptions As New ExecutionOptions

    'creates a new DaSession object and adds it to the application
    Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

    'creates a new DaSubscription objects and adds it to the created session
    Dim subscription As New DaSubscription ( Convert.ToUInt32(1000), session)

    'creates a new DaItem object and adds it to the created subscription
    Dim item1 As New DaItem("maths.sin" , subscription)

    'connects session, subscription and item1 to the server
    session.Connect(True, False, executionOptions)

    'activates the session, subscription and item1      
    session.Connect(True, True, executionOptions)

    'disconnects session, subscription and item1 from the server
    session.Disconnect(executionOptions)
    ...
End Sub
```
### C++
```
int _tmain(int argc, _TCHAR* argv[])
{
    Application* application = getApplication();

    application->initialize();

    // creates a new DaSession object and adds it to the application
    DaSession* session = new DaSession("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}");

    // creates a new DaSubscription object and adds it to the created session
    DaSubscription* subscription = new DaSubscription(1000, session);    

    // creates a new DaItem object and adds it to the created subscription
    DaItem* item1 = new DaItem(_T("maths.sin"), subscription);

    ExecutionOptions executionOptions;
    // connects session, subscription and item1 to the server
    session->connect(TRUE, FALSE, &executionOptions);

    // activates the session, subscription and item1
    session->connect(TRUE, TRUE, &executionOptions);

    // disconnects session, subscription and item1 from the server
    session->disconnect(NULL);

    application->terminate();            
    releaseApplication();
    delete session;
    delete subscription;
    delete item1;
}
```
### Example 4

The following example demonstrates how to deeply connect, activate and then disconnect a DaSession object. The operations are done in an **asynchronous** way. For a more detailed example on this topic check the **DAConsole** example.

### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...
public static void HandleStateChanged(ObjectSpaceElement sender, EnumObjectState state)
{
    if ((sender is DaSession) || (sender is DaSubscription))
    {                
        Console.WriteLine( sender + " State Changed - " + state);
    }
    if (sender is DaItem)
    {
        DaItem item = sender as DaItem;
        Console.WriteLine( item + " " + item.Id + " State Changed - " + state);
    }
}

public static void HandlePerformStateTransition(
    ObjectSpaceElement sender,
    uint executionContext,
    int result)
{            
    if(ResultCode.SUCCEEDED(result)){
        if ((sender is DaSession) || (sender is DaSubscription))
        {                
            Console.WriteLine( sender + " Performed state transition - "  + executionContext );
        }
        if (sender is DaItem)
        {
            DaItem item = sender as DaItem;
            Console.WriteLine( sender + " " + item.Id + " Performed state transition - "  + executionContext );
        }
    }
    else{
        Console.WriteLine("  Performed state transition failed");
    }
}

static void Main(string[] args) {

    // gets the application instance
    Application application = Application.Instance;

    ExecutionOptions executionOptions = new ExecutionOptions();
    executionOptions.ExecutionType = EnumExecutionType.ASYNCHRONOUS;
    executionOptions.ExecutionContext = 0;

    // creates a new DaSession object and adds it to the application
    DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

    session.StateChangeCompleted += new StateChangeEventHandler(HandleStateChanged);
    session.PerformStateTransitionCompleted += new PerformStateTransitionEventHandler (HandlePerformStateTransition);

    // creates a new DaSubscription object and adds it to the created session
    DaSubscription subscription = new DaSubscription(500, session);

    subscription.StateChangeCompleted += new StateChangeEventHandler(HandleStateChanged);
    subscription.PerformStateTransitionCompleted += new PerformStateTransitionEventHandler (HandlePerformStateTransition);

    // creates a new DaItem object and adds it to the created subscription
    DaItem item1 = new DaItem("maths.sin",subscription);

    item1.StateChangeCompleted += new StateChangeEventHandler(HandleStateChanged);
    item1.PerformStateTransitionCompleted += new PerformStateTransitionEventHandler (HandlePerformStateTransition);

    // connects session, subscription and item1 to the server
    session.Connect(true, false, executionOptions);
    executionOptions.ExecutionContext++;

    // activates the session, subscription and item1
    session.Connect(true, true, executionOptions);
    executionOptions.ExecutionContext++;

    // disconnects session, subscription and item1 from the server
    session.Disconnect(executionOptions);
    executionOptions.ExecutionContext++;

    application.Terminate();
}
```
### VB .NET
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...

Public Shared Sub HandleStateChanged(ByVal obj As ObjectSpaceElement, ByVal state As EnumObjectState)
    If Not Type.ReferenceEquals(obj.[GetType](), GetType(DaItem)) Then
        System.Console.WriteLine(Convert.ToString(obj) + " State Changed - " + Convert.ToString(state))
    Else
        Dim item As DaItem
        item = DirectCast(obj, DaItem)
        System.Console.WriteLine(Convert.ToString(obj) + " " + item.Id + " State Changed - " + Convert.ToString(state))
    End If
End Sub

Public Shared Sub HandlePerformStateTransition(ByVal obj As ObjectSpaceElement, ByVal executionContext As System.UInt32, ByVal result As System.Int32)

    If ResultCode.SUCCEEDED(result) Then
        If Not Type.ReferenceEquals(obj.[GetType](), GetType(DaItem)) Then
            System.Console.WriteLine(Convert.ToString(obj) + " Performed state transition - " + Convert.ToString(executionContext))
        Else
            Dim item As DaItem
            item = DirectCast(obj, DaItem)
            System.Console.WriteLine(Convert.ToString(obj) + " " + item.Id + " Performed state transition -" + Convert.ToString(executionContext))
        End If
    Else
        System.Console.WriteLine("  Performed state transition failed")
    End If
End Sub
...
Shared Sub Main(ByVal args As String())

    'gets the application instance
    Dim application As Application = Application.Instance

    application.Initialize()

    'set the execution options to asynchronous
    Dim executionOptions As New ExecutionOptions
    executionOptions.ExecutionType = EnumExecutionType.ASYNCHRONOUS
    executionOptions.ExecutionContext = Convert.ToUInt32(0)

    'creates a new DaSession object and adds it to the application
    Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

    AddHandler session.StateChangeCompleted, AddressOf HandleStateChanged
    AddHandler session.PerformStateTransitionCompleted, AddressOf HandlePerformStateTransition

    'creates a new DaSubscription objects and adds it to the created session
    Dim subscription As New DaSubscription ( Convert.ToUInt32(1000), session)

    AddHandler subscription.StateChangeCompleted, AddressOf HandleStateChanged
    AddHandler subscription.PerformStateTransitionCompleted, AddressOf HandlePerformStateTransition

    'creates a new DaItem object and adds it to the created subscription
    Dim item1 As New DaItem("maths.sin", subscription)

    AddHandler item1.StateChangeCompleted, AddressOf HandleStateChanged
    AddHandler item1.PerformStateTransitionCompleted, AddressOf HandlePerformStateTransition

    'connects session, subscription and item1 to the server
    session.Connect(True, False, executionOptions)
    Decimal.op_Increment(Convert.ToDecimal(executionOptions.ExecutionContext))

    'activates the session, subscription and item1
    session.Connect(True, True, executionOptions)
    Decimal.op_Increment(Convert.ToDecimal(executionOptions.ExecutionContext))

    'disconnects session, subscription and item1 from the server
    session.Disconnect(executionOptions)
    Decimal.op_Increment(Convert.ToDecimal(executionOptions.ExecutionContext))

    application.Terminate();
    ...
End Sub
```
### C++
```
class MyDaSession : public DaSession{

    public:

    MyDaSession (const tstring& url): DaSession(url) {}
    void handleStateChangeCompleted(EnumObjectState state){

        _tprintf(_T("\n%s state changed - %s\n"), "Session",state);
    }

    void handlePerformStateTransitionCompleted(
        DWORD executionContext,
        LONG result)
    {
        _tprintf("\n");
        if (SUCCEEDED(result)){
            _tprintf(_T("%s performed state transition - context:%lu\n"), "Session", executionContext);
        } // end if
        else{
            _tprintf(" performed state transition failed");
        }
    } // end handlePerformStateTransitionCompleted
}

int _tmain(int argc, _TCHAR* argv[])
{

    ExecutionOptions executionOptions;

    executionOptions.setExecutionType(ASYNCHRONOUS);    
    executionOptions.setExecutionContext(0);
    executionContext = executionOptions.getExecutionContext();

    Application* application = getApplication();
    application->initialize();

    // creates a new MyDaSession object and adds it to the application
    MyDaSession* session = new MyDaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}"));    

    // creates a new DaSubscription object and adds it to the created session
    DaSubscription* subscription = new DaSubscription(1000, session);    

    // creates a new DaItem object and adds it to the created subscription
    DaItem* item1 = new DaItem(_T("maths.sin"), subscription);

    // connects session, subscription and item1 to the server
    session->connect(TRUE, FALSE, &executionOptions);
    executionOptions.setExecutionContext(++executionContext);

    // activates the session, subscription and item1
    session->connect(TRUE, TRUE, &executionOptions);    
    executionOptions.setExecutionContext(++executionContext);

    // disconnects session, subscription and item1 from the server
    session->disconnect(&executionOptions);    
    executionOptions.setExecutionContext(++executionContext);

    application->terminate();            
    releaseApplication();
    delete session;
}
```
### Note
```
A session object that is in the activated state doesn’t pass directly from activated to disconnected state. It first 
passes to connected and afterwards to disconnected state.
```
### Note
```
In an asynchronous scenario, if the current state and target state are different, the state transition is performed 
and the user is notified about the state change. If the current state and target state are the same no message 
about the state change will be sent, because in this case the object’s state remains the same.
