## **Update object attributes**

The **DaSession**, **AeSession**, **DaSubscription**, **AeSubscription** and **DaItem** objects are the equivalents in the Softing OPC Toolkit of the OPC objects. They have attributes values which can be of great interest for the user. The Softing OPC Toolkit provides a very simple way of finding or establishing new values for the attributes of these objects, by using properties (.NET) or “get” and “set” methods (C++).

When the value of an attribute is to be found, first the server should be asked to give the most recent value of that attribute. This is done by updating the values from the server (using the **GetAttributesFromServer** method). After the update, the value can be gotten using the corresponding property (.NET) or a “get” method (C++).

The same, when a new value for an attribute is to be established, the new value must be set using the corresponding property (.NET) or a “set” method (C++). Afterwards the changes must be transmitted to the server. This is done by updating the values to the server (using the **SetAttributesToServer** method).

So, in both cases there must be an interaction between the client and the server. This interaction can be done with or without the user to be aware of it.

The toolkit provides two possibilities of updating the value of an object’s attributes. These possibilities are modeled using the **UpdateAttributesBehavior** property (.NET) or **setUpdateAttributesBehavior** method (C++) of the Application class. Using them makes possible to specify if the update of the attributes is explicit or implicit.

With an **implicit** update a user doesn’t have to mind about updating the attributes values from or to the server. For every attribute value that needs to be gotten, first a call of updating attributes from the server is automatically done, without the user to explicitly specify this, and only after the attributes values are really returned. For every new attribute value that needs to be established, after the user sets the new value, an update of the attribute to the server is automatically performed.

### Note
```
The implicit update is simple to use but slows down the client because the update is done for each attribute separately!
```
Therefore the explicit update comes to offer an alternative. In this case the user must explicitly call the **GetAttributesFromServer** and **SetAttributesToServer** methods, not for every attribute but for a set of attributes.

First an update from the server must be done for all the attributes that are considered to be of interest. Then the attributes’ values can be gotten. If new values are wanted to be set for some attributes, then the new values will be first set and then an update to the server will be explicitly made for all the new values.

The examples below show how to update attributes values both in implicit and explicit way.

###  Example 1

The following example demonstrates how to set the attributes of a DaSession object using the **implicit** way of updating attributes.

For a more detailed example on this topic check the **DAObjectAttributes** example.

### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...
// gets the application instance
Application application = Application.Instance;

// initializes the application
application.Initialize();

// creates a new DaSession object
DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}");

// chooses to implicit update the session’s attributes
Application.Instance.UpdateAttributesBehavior = EnumUpdateAttributesBehavior.IMPLICIT_UPDATE;

// Sets new values for the session’s attributes
session.ClientName = "Update attributes example";
session.LocaleId = "de-DE";

// Displays the new value of the session attributes
Console.WriteLine("Session Attributes:");                    
Console.WriteLine("URL: " + session.Url);    
Console.WriteLine("ClientName: " + session.ClientName);
Console.WriteLine("LocaleID: " + session.LocaleId);
...


//The result written on the console is:
//Session Attributes:
    //URL: "opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}"
    //ClientName: Update attributes example
    //LocaleID: de-De
```
### VB .NET
```
Imports Softing.OPCToolbox
Imports Softing.OPCToolbox.Client
...
'gets the application instance
Dim application As Application = Application.Instance

'initializes the application
application.Initialize()

'creates a new DaSession object
Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

'chooses to implicit update the session’s attributes
application.UpdateAttributesBehavior = EnumUpdateAttributesBehavior.IMPLICIT_UPDATE

'sets new values for the session’s attributes
session.ClientName = "Update attributes example"
session.LocaleId = "de-DE"

'displays the new value of the session attributes
Console.WriteLine("Session Attributes:")                    
Console.WriteLine("URL: " + session.Url)    
Console.WriteLine("ClientName: " + session.ClientName)
Console.WriteLine("LocaleID: " + session.LocaleId)
...
```

### C++
```
using namespace SoftingOPCToolboxClient;
...
Application* application = getApplication();
application.Initialize();

DaSession* session = new DaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}");    

// Chooses to implicit update the session’s attributes
application->setUpdateAttributesBehavior(EnumUpdateAttributesBehavior_IMPLICIT_UPDATE);

// Sets new values for the session’s attributes
session->setClientName(_T("Update attributes example"));            
session->setLocaleId(_T("de-de"));

// Displays the new values of the session attributes
_tprintf(_T("\nSession Attributes:\n"));
_tprintf(_T("  URL: %s\n"), session->getUrl().c_str());
_tprintf(_T("  Client Name: %s\n"), session->getClientName().c_str());
_tprintf(_T("  LCID: %lu\n"), session->getLocaleId());    
...
```

### Example 2
The following example demonstrates how to set the attributes of a DaSession object using the **explicit** way of updating attributes.

For a more detailed example on this topic check the **DAObjectAttributes** example.

### C#
```
using Softing.OPCToolbox.Client;
using Softing.OPCToolbox;
...
// gets the application instance
Application application = Application.Instance;

// initializes application
application.Initialize();

// creates a new Session object
DaSession session = new DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}",session);

// chooses to explicit update the session’s attributes
Application.Instance.UpdateAttributesBehavior = EnumUpdateAttributesBehavior.EXPLICIT_UPDATE;

// sets new values for the session’s attributes
session.ClientName = "Update attributes example";                
session.LocaleId = "de-DE";        

// updates the values of all modified attributes to the server        
session.SetAttributesToServer(
    new EnumObjectAttribute[] {EnumObjectAttribute.DASESSION_CLIENTNAME,EnumObjectAttribute.DASESSION_LCID},
    out updateResults,
    new ExecutionOptions());
...

// updates the values of all attributes from the server
session.GetAttributesFromServer(
    new EnumObjectAttribute[] {EnumObjectAttribute.AESESSION_ALL},
    out updateResults,
    new ExecutionOptions());

// displays the new value of the session attributes
Console.WriteLine("Session Attributes:");            
Console.WriteLine("URL: " + session.Url);    
Console.WriteLine("ClientName: " + session.ClientName);    
Console.WriteLine("LocaleID: " + session.LocaleId);
...

//The result written on the console is:
//Session Attributes:
    //URL: "opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}"
    //ClientName: Update attributes example
    //LocaleID: de-De
```

### VB .NET
```
Imports Softing.OPCToolbox
Imports Softing.OPCToolbox.Client
...
'gets the application instance
Dim application As Application = Application.Instance

'initializes the application
application.Initialize()

'creates a new DaSession object
Dim session As New DaSession("opcda://localhost/Softing.OPCToolboxDemo_ServerDA/{2e565242-b238-11d3-842d-0008c779d775}")

'chooses to implicit update the session’s attributes
application.UpdateAttributesBehavior = EnumUpdateAttributesBehavior.EXPLICIT_UPDATE

'Sets new values for the session’s attributes
session->setClientName(_T("Update attributes example"));            
session->setLocaleId(_T("de-de"));    

sessionLcid = [Enum].ToObject(GetType(EnumObjectAttribute), EnumObjectAttribute.DASESSION_LCID)
clientName = [Enum].ToObject(GetType(EnumObjectAttribute), EnumObjectAttribute.DASESSION_CLIENTNAME)
Dim results As Integer() = Nothing
Dim executionOptions As New ExecutionOptions()

'updates the values of all modified attributes to the server        
session.session.SetAttributesToServer(
    New EnumObjectAttribute() {sessionLcid, clientName}, _
    results, _
    executionOptions)        
...

'updates the values of all attributes from the server
session.session.GetAttributesFromServer(
    New EnumObjectAttribute() {EnumObjectAttribute.AESESSION_ALL}, _
    results, _
    executionOptions)        

'displays the new value of the session attributes
Console.WriteLine("Session Attributes:")        
Console.WriteLine("URL: " + session.Url)    
Console.WriteLine("ClientName: " + session.ClientName)
Console.WriteLine("LocaleID: " + session.LocaleId)
...
```
### C++
```
using namespace SoftingOPCToolboxClient;
...
Application* application = getApplication();
application->initialize();

DaSession* session = new DaSession(_T("opcda:///Softing.OPCToolboxDemo_ServerDA.1/{2E565242-B238-11D3-842D-0008C779D775}"));

// chooses to explicit update the session’s attributes
application->setUpdateAttributesBehavior(EnumUpdateAttributesBehavior_EXPLICIT_UPDATE);

unsigned long updateCount = 2;    
std::vector<EnumObjectAttribute> whatAttributes;
std::vector<LONG> updateResults(updateCount, E_FAIL);

whatAttributes.push_back(EnumObjectAttribute_DASESSION_CLIENTNAME);
whatAttributes.push_back(EnumObjectAttribute_DASESSION_LCID);

// sets new values for the session’s attributes
session->setClientName(_T("Update attributes example"));            
session->setLocaleId(_T("de-de"));

// updates the values of all modified attributes to the server
session->setAttributesToServer(
    whatAttributes,
    updateResults,
    NULL);
...

// updates the values of all attributes from the server
session->getAttributesFromServer(
    whatAttributes,
    updateResults,
    NULL);

// displays the new value of the session attributes
_tprintf(_T("\nSession Attributes:\n"));
-Tprintf(_T("  URL: %s\n"), session->getUrl().c_str());
_tprintf(_T("  Client Name: %s\n"), session->getClientName().c_str());
_tprintf(_T("  LCID: %lu\n"), session->getLocaleId());    
...
```

The GetAttributesFromServer and SetAttributesToServer methods can be used only in case of session (DaSession, AeSession) or subscription (DaSubscription, AeSubscription) objects. In case the update must be done for DaItem objects' attributes, the **GetDaItemAttributesFromServer** and **SetDaItemAttributesToServer** methods must be used.

For a detailed example on this topic check the **DaObjectAttributes** and **AEObjectAttributes** examples.