## **How to Build Server's Address Space**

### How to build address space

Every OPC Server programmed with the OPC Development Toolkit must provide an OPC DA respective AE Addresss Space. An Address space element must be created prior to be used within a transaction and used by an OPC client.

The Address Space of the Toolkit is stored in a tree without a depth limitation. Both the branches and the leaves of the tree are represented by address space elements. They are differentiated by hasChildren property. An address space element may have I/O regardles of its branch or leaf nature. Furthermore a DA Address Space Element may have properties regardles if it is a branch or node.

It is possible to create either a flat name space (only Address space elements under the root element) or a hierarchical name space.

A leaf Address Space Element always represents a process value that can be read or written using OPC. Branch Address Space Elements are used to structure the process values in the name space. Properties offer descriptive information for Address Space Elements. Since the elements of the address space tree are created by the server application, they can be derived from the base clases DaAddresssSpaceElement and DaProperty.

The Toolkit uses the OPC address space tree to:

-   process calls of the OPC Clients for browsing of the name space
-   relate item objects to process variables ( Address space elements with I/O enabled)

The 'fully qualified item ID' represents the path to an address space element in the address space. The Toolkit supports two ways of providing the name space:

Both OPC DA and OPC AE Servers may have either Object-based or String-based address spaces. An object-based address space (static address space) is created in the server application by instantiating the AddressSpaceElement (or any inheriting) class and adding these elements to the root / other nodes. The string-based address space is considered to be a dynamic address space as it is entirely handled in the customer application by returning requested values to the callbacks the Toolkit CORE invokes during the runtime of the OPC Server.

Object-based address provides a much faster access time than string-based address space. Given the callback invoking and the complexity of such processing amd information marshalling tasks, the string-based namespace is by definition slower than the object-based address space.

**Object-based address space**

The server application provides the object-based address space via the creation of an object tree in the Toolkit. All actions related to the object-based address space use the objects of the address space tree. An object-based address space is also called static, because it is being created at server startup. Normally it does not change at runtime.

**String-based address space**

The string-based address space exists in string form only until an OPC Client creates an item object which uses an element of the string-based name space, upon which an object for this address space element is generated. As soon as no OPC Client uses this object it is released. A string-based address space is therefore also called dynamic address space, because the according objects are being created and released dynamically. If a string-based address space follows a special syntax (e.g. S7 syntax: DB31.DBB2) then it is called syntax-based address space.

**Both above**

An OPC Server application can support both mechanisms at the same time.

The user must specify the type of address space used at with the construction of the Address space root overriding the Creator.CreateDaAddressSpaceRoot() method.

###  Object-Based Name Space

The object-based name space is held in the name space object tree. The OPC Server application queries the root of the address space tree from the Toolkit and generates the objects of the name space below it from the configuration of the server application. For this purpose, it uses the common tree management methods of the Address Space Element class. At all times, objects of the name space tree can be accessed using the children lists. The user is engouraged though to keep the own references to the address space elements accordingly to the server's own particularities.

-   Advantages : All actions related to the name space can be carried out entirely by the Toolkit.
    
-   Disadvantages : Memory space is occupied by objects that will possibly never be used by an OPC Client.
    

###  String-Based Name Space

The objects of the string-based name space are not created until the Client adds an item object that uses them. When a Address Space Element for a process value of the string-based name space has been created, it becomes part of the object-based name space tree for its lifetime. The object is deleted after the last external reference (from outside the address space space tree) to a dynamically created tag has been released.

All actions applied for the dynamic name space use a string with the complete path of the name space element. The OPC Server application must implement methods which determine whether a string describes a valid address space element or property. This way, it can be verified during the creation and validation of item objects whether the passed item ID is valid or not.

The Toolkit also supports browsing the string-based name space. For this purpose, the server application must implement additional methods to determine the children of an address space element and the properties of a string-based name space element. In addition, for the browsing of string-based tag objects, the data type and the access paths must be announced.

The methods that must be overriden in order to accomplish this are:

DaAddressSpaceRoot.QueryAddressSpaceElementData();

DaAddressSpaceRoot.QueryAddressSpaceElementDataEx();

DaAddressSpaceRoot.QueryAddressSpaceElementChildren();

DaAddressSpaceRoot.QueryAddressSpaceElementProperties();

-   Note : If QueryAddressSpaceElementDataEx callback is fully implemented, QueryAddressSpaceElementData callback will no longer be invoked. Yet, if both callbacks are implemented, both callbacks will be invoked for a non-existing string-based address space elements (having that QueryAddressSpaceElementDataEx callback will return an error, as the string-based address space elements do not exist).
-   Advantages : Only those name space objects that are actually used exist. An address space element that defines process values based on a syntax can be implemented (E.g. Dev1Off18TypeI4 stands for an integer value of device 1 at offset 18).
-   Disadvantages : The server application must implement the virtual methods for providing the string-based name space.

Usage of string based addresss space can be found in the 'server\AddressSpace' sample.

### Example

The folowing example shows how to build an **Object-Based Name Space** server. The BuildAddressSpace function must be called after the server has been successfully initialized.
### C#
```
using Softing.OPCToolbox.Server;
using Softing.OPCToolbox;
...
static public MyDaAddressSpaceElement MyCountTag;
static public MyDaAddressSpaceElement randTag1s;

public int BuildAddressSpace()
{
    try
    {
        MyDaAddressSpaceElement  node;

        node = new MyDaAddressSpaceElement();
        node.Name = "MyNode";
        node.IoMode = EnumIoMode.NONE;
        node.HasChildren = true;
        Application.Instance.DaAddressSpaceRoot.AddChild(node);

        MyCountTag = new MyDaAddressSpaceElement();
        MyCountTag.Name = "counter";
        MyCountTag.AccessRights = EnumAccessRights.READABLE;
        MyCountTag.Datatype =  typeof(double);
        MyCountTag.IoMode = EnumIoMode.POLL;
        node.AddChild(MyCountTag);

        DaProperty MyProperty = new DaProperty();            
        MyProperty.Id = 101;
        MyProperty.Name = "Counter property";
        MyProperty.Description = "Description";
        MyProperty.ItemId = MyProperty.Name;
        MyProperty.Datatype = typeof(String);
        MyProperty.AccessRights = EnumAccessRights.READWRITEABLE;
        MyCountTag.AddProperty(MyProperty);

        node = new MyDaAddressSpaceElement();
        node.Name = "random";
        node.IoMode = EnumIoMode.POLL;
        node.HasChildren = true;
        Application.Instance.DaAddressSpaceRoot.AddChild(node);

        randTag1s = new MyDaAddressSpaceElement();
        randTag1s.Name = "1s";
        randTag1s.AccessRights = EnumAccessRights.READWRITEABLE;
        randTag1s.Datatype = typeof(Int32);
        randTag1s.IoMode = EnumIoMode.REPORT;
        node.AddChild(randTag1s);

    }
    catch(Exception exc)
    {
        Trace(
            EnumTraceLevel.ERR,
            EnumTraceGroup.USER1,
            "OpcServer:BuildAddressSpace",
            exc.ToString());
        return (int)EnumResultCode.E_FAIL;
    }    //    end try...catch

    return (int)EnumResultCode.S_OK;
}    //    end BuildAddressSpace
```
### VB .NET
```
Imports Softing.OPCToolbox.Server
Imports Softing.OPCToolbox
...
Public Shared randTag1s As MyDaAddressSpaceElement
Public Shared MyCountTag As MyDaAddressSpaceElement

Public Function BuildAddressSpace() As Integer
    Try
        Dim node As MyDaAddressSpaceElement

        node = New MyDaAddressSpaceElement()
        node.Name = "MyNode"
        node.IoMode = EnumIoMode.NONE
        node.HasChildren = True
        Application.Instance.DaAddressSpaceRoot.AddChild(node)

        MyCountTag = New MyDaAddressSpaceElement()
        MyCountTag.Name = "counter"
        MyCountTag.AccessRights = EnumAccessRights.READABLE
        MyCountTag.Datatype = GetType(Double)
        MyCountTag.IoMode = EnumIoMode.POLL
        node.AddChild(MyCountTag)

        Dim MyProperty As New DaProperty()

        MyProperty.Id = 101
        MyProperty.Name = "Counter property"
        MyProperty.Description = "Description"
        MyProperty.ItemId = MyProperty.Name
        MyProperty.Datatype = GetType(String)
        MyProperty.AccessRights = EnumAccessRights.READWRITEABLE
        MyCountTag.AddProperty(MyProperty)

        node = New MyDaAddressSpaceElement()
        node.Name = "random"
        node.IoMode = EnumIoMode.POLL
        node.HasChildren = True
        Application.Instance.DaAddressSpaceRoot.AddChild(node)

        randTag1s = New MyDaAddressSpaceElement()
        randTag1s.Name = "1s"
        randTag1s.AccessRights = EnumAccessRights.READWRITEABLE
        randTag1s.Datatype = GetType(Integer)
        randTag1s.IoMode = EnumIoMode.REPORT
        node.AddChild(randTag1s)

    Catch exc As Exception
        Trace(EnumTraceLevel.ERR, m_traceGroupUser1, "OpcServer:BuildAddressSpace", exc.ToString())
        Return EnumResultCode.E_FAIL
    End Try

    Return EnumResultCode.S_OK
End Function
```
### C++
```
using namespace SoftingOPCToolboxClient;
...
MyDaAddressSpaceElement* MyCountTag;
MyDaAddressSpaceElement* randTag1s;

long OpcServer::buildDaNameSpace(void)
{
    MyDaAddressSpaceElement *node;
    tstring aName;

    DaAddressSpaceRoot* daRoot = getApp()->getDaAddressSpaceRoot();

    node = new MyDaAddressSpaceElement();
    aName = tstring(_T("MyNode"));
    node->setName(aName);
    node->setIoMode(EnumIoMode_NONE);
    node->setHasChildren(true);
    daRoot->addChild(node);

    MyCountTag = new MyDaAddressSpaceElement();
    aName = tstring(_T("counter"));            
    MyCountTag->setName(aName);
    MyCountTag->setAccessRights(EnumAccessRights_READABLE);
    MyCountTag->setDatatype(VT_R8);
    MyCountTag->setIoMode(EnumIoMode_POLL);
    node->addChild(MyCountTag);

    DaProperty* MyProperty = new DaProperty();
    MyProperty->setId(101);
    tstring propName(_T("Counter property"));
    MyProperty->setName(propName);
    tstring propDescription(_T("Description"));
    MyProperty->setDescription(propDescription);
    MyProperty->setItemId(propName);
    MyProperty->setDatatype(VT_I2);
    MyProperty->setAccessRights(EnumAccessRights_READWRITEABLE);
    MyCountTag->addProperty(MyProperty);

    node = new MyDaAddressSpaceElement();
    aName = tstring(_T("random"));
    node->setName(aName);
    node->setIoMode(EnumIoMode_POLL);
    node->setHasChildren(true);
    daRoot->addChild(node);


    randTag1s = new MyDaAddressSpaceElement();
    aName = tstring(_T("1s"));
    randTag1s->setName(aName);
    randTag1s->setAccessRights(EnumAccessRights_READWRITEABLE);
    randTag1s->setDatatype(VT_I4);
    randTag1s->setIoMode(EnumIoMode_REPORT);
    node->addChild(randTag1s);
    return S_OK;
}   //  end buildDaNameSpace
```